//
//  HFRepresenterVerticalScroller.h
//  HexFiend_2
//
//  Copyright 2007 ridiculous_fish. All rights reserved.
//

#import <HexFiend/HFRepresenter.h>

/*! @class HFVerticalScrollerRepresenter
    @brief An HFRepresenter responsible for showing a vertical scroll bar.
    
    HFVerticalScrollerRepresenter is an HFRepresenter whose view is a vertical NSScroller, that represents the current position within an HFController "document."  It has no methods beyond those of HFRepresenter.
    
    As HFVerticalScrollerRepresenter is an especially simple representer, it makes for good sample code.
*/
@interface HFVerticalScrollerRepresenter : HFRepresenter {

}

@end
